/** \file emscher.c
 *  An application for easyAVR (7), Arduino and other supported AVR platforms
 *
 *  This simple demo application may use the same USART / serial link as the
 *  serial bootloader to exchange textual messages with a very simple serial
 *  terminal program like vsst.exe or a more comfortable one like HTerm.
 *
 *  Besides that it is nearly empty (does nothing complicated), it may serve
 *  as starting point for own developments. Clean and build by:
 *  make MAIN_F=emscher clean all
 *
 *  <br /> &nbsp;
 * \ref copyLicNote " Copyright &copy; " 2014 Albrecht Weinert, Bochum
 * \author  Albrecht Weinert  &nbsp; <<a href="http://a-weinert.de">a-weinert.de</a>>
 * 1
 * 25.01.2017
 */
// Copyright 2014, 2015 Albrecht Weinert, Bochum

#include "arch/expConf.h"
// next 3 includes point to elements of weAutSys' serial bootloader project
#include "arch/config.h"
#include "bootloader/boot109.h"
#include "we-aut_sys/ll_common.h"

#include <avr/pgmspace.h>
// #include <avr/interrupt.h>
#include <stddef.h>

uint8_t resetCauses;


/** main's welcome greeting and Copyright */
INFLASH(char const mainWlc[]) = "\n\n\n\n\n\n\n\n\n"
      " * * * *   AVR demo program emscher V0.2              * * * *  \n"
      "                                                               \n"
      " It links and uses weAutSys' serial bootloader functions.      \n"
      " Copyright (c) 2014  weinert - automation      <weAut.de>      \n"
      " Prof. Dr.-Ing. Albrecht Weinert,  Bochum  <a-weinert.de>      \n"
      "                                                               \n";

// may be commented out / in to demonstrate the improvement
// int main(void) __attribute__((OS_main));  // OS_main is like ((noreturn))

/** The application (system) program start
 *
 *  This is the normal system program's program entry.
 *
 *  At present it is just a bit serial (HMI) communication
 *  and an output of left button (PC0) to left gn LED (/PB7).
 *  Scan interval for the button input is as fast as possible. Alive sign
 *  is  by blinking the right rd LED (/PB0; ports for PLATFORM=boAVR2013).
 *
 *  This is done for PLATFORM=boAVR2013 and PLATFORM=arduMega2650 only. On
 *  other platforms we just echo the serial input or send a ~ (Tilde) every
 *  25s to show being alive.
 */
int main(void){
   // The bootloader leaves reset causes intact if it just returns by
   // timeout. In that case it wasn't really active as bootloader
   // (communicating with AVRdude e.g.).
   resetCauses = MCUSR; MCUSR = 0; // save & clear reset cause
   ///  basicSystemInit();  // done in boot loader, already

   sendSerBytes_P(LOW_ADD(mainWlc));  // a greeting (low flash)
   sendSerBytes_P(FAR_ADD(bootloaderPlatf));  // platform info by bootloader

   sendSerBytes_P(FAR_ADD(bootTextReset));  // bootloader function and text
   sendSerBytes_P(resetCauseText_P(resetCauses));
   sendSerByte('\n'); sendSerByte('\n');

#ifdef     boAVR2013
   // make by: make PLATFORM=boAVR2013 MAIN_F=emscher PRG_APP_P=com4 clean all
   // flash by: make PLATFORM=boAVR2013 MAIN_F=emscher PRG_APP_P=com4 progapp
   // left button (PC0) to left gn LED (/PB7).
   // live LED right rd LED (/PB0)
# define MAINLED_PORT  PORTB
# define MAINLED_MASK  0x80
//  # define LIVELED_PORT  PORTB
//  # define LIVELED_MASK  0x01
# define BUTTON_PORT  PINC
# define BUTTON_MASK  0x01

   DDRB |= (LIVELED_MASK|MAINLED_MASK); // outer LEDs output (& by bootloader)
   DDRC = 0; // C all input (buttons) (already by bootloader)
   PORTC = 0xFF; // all pull-ups on  (already by bootloader)


#elif defined(arduMega2560)
   // make by: make PLATFORM=arduMega2560 MAIN_F=emscher PRG_APP_P=com5 clean all
   // flash by: make PLATFORM=arduMega2560 MAIN_F=emscher PRG_APP_P=com5 progapp

   // PF1 (A1) button    (hi-active if "brick buttons used)
   // PB4 (D10) main LED   (hi-active)
   //    out  xxxxx PB7 live LED (buildin, ye, hi-active)
   // PH6 (D9) live LED (hi-active)

# define MAINLED_PORT  PORTB
# define MAINLED_MASK  0x10
# define BUTTON_PORT  PINF
# define BUTTON_MASK  0x02

   DDRB |= MAINLED_MASK; // LED out output (already by bootloader ?)
   DDRF = 0; // F all input (for button) (already by bootloader)
   PORTF = 0xFF; // F all pull-ups on
   DDRH |= LIVELED_MASK; // H6 out live LED


#else
   // dummy port defs, not used here
# define MAINLED_PORT  PORTB
# define MAINLED_MASK  0x10
# ifndef LIVELED_PORT
#  define LIVELED_PORT  PORTB
#  define LIVELED_MASK  0x40
# endif
# define BUTTON_PORT  PINB
# define BUTTON_MASK  0x02

   for(;;) { // endless loop
       if (waitSerByte(0)) { // timeout while waiting 25.6s
           sendSerByte('~'); // yes we can (one AC per 25 s timeout)
           continue; // stay in endless loop
       } // timeout while waiting for UART byte
       uint8_t val = getSerByte(); // Get the byte received (one is ready)
       sendSerByte(val); // All character are echoed by this "application"
   } // endless for
#endif

   for(uint16_t i=0;;++i) { // endless loop
      uint8_t inBut = BUTTON_PORT;
      if (inBut & BUTTON_MASK) { // button input hi
         MAINLED_PORT |= MAINLED_MASK;
      } else { // button hi else lo
         MAINLED_PORT &= ~MAINLED_MASK;
      }  // button  lo
      if (i & 0x8000) { // blink interval bit
         orPort(LIVELED_PORT, LIVELED_MASK);
         // MAINLED_PORT &= ~MAINLED_MASK; // test only
      } else { // button released else pressed
         andPort(LIVELED_PORT, ~LIVELED_MASK);
         // MAINLED_PORT |= MAINLED_MASK; // test only
      }  // button  pressed
   } // endless for
} // main()   ("emscher")


