/** \file lcd1602b.c
 *
 *  Implementation of LCD module (1602B types) functions
 *
 *  Copyright (c) 2014, 2015  Albrecht Weinert, Bochum
 *
 *  Currently we support three hardware target platforms (PLATFORM=..): <br />
  - easyAVR32
  - arduMega2560
  - boAVR2013
  - arduinoUno
 *
 *  All else see in the include file lcd1602b.h.
 *
 *  1
 *  25.01.2017
 */


#include "lcd1602b.h"
#include "arch/expConf.h"
#include "we-aut_sys/ll_common.h"  // just to please Eclipse
#include "bootloader/boot109.h"


//---  timing helper  (all using spin waiting as yet)     -------------------

/*  A pre-initialisation pause of 20ms or 30 ms
 *
 *  This function should be called before \ref initLCD if the LCD module is
 *  to be initialised shortly after power up.
 *
 *  See all else in \ref pause4LCDmodule (in the declaring .h file).
 *
 *  \param sup3V 3: the supply is only 3V: the longer wait is needed
 */
void preInitLCDpause(uint8_t sup3V){
   uint16_t cnt = sup3V == 3 ? (30000 / 25) : (20000 / 25);
   do { wait25(); } while (--cnt);
} // preInitLCDpause(uint8_t)


/*  Read the address and busy flag
 *
 *  This function does nothing and returns 0xFF if the R/W on the TC1602B is
 *  \ref  RW_NOTFIXED "fixed" to Gnd.
 *
 *  This implementation is still under development. Due to a lack of
 *  suitable platforms it was tested only for 4-bit mode and only in the
 *  brick pervert variant, yet. <br />
 *  Background: On most platforms the R/W is just  soldered to Gnd.
 */
uint8_t addressFromLCD(){
#if  RW_NOTFIXED
   uint8_t addr = 0;
   setLCDdataPorts4in();
   andPort(LCD_RS_PORT, ~LCD_RS_MASK); // RS  set 0
#if LCD_DATA_USE == BRICK_PERVERT
     orPort(LCD_E_PORT, LCD_E_MASK);   // Enable  set 1;
     asm volatile ("nop");  // internal delay
     asm volatile ("nop");  // data setup
     uint8_t in1 = inPort(LCD_D4_PORT);  // in D4 port nib bit 0
     uint8_t in2 = brickPervertAlignR(inPort(LCD_DATA_PORT));
     addr = (in2 <<4) & 0xF0;
     andPort(LCD_E_PORT, ~LCD_E_MASK); // Enable  set 0;
     if (in1 & LCD_D4_MASK) addr |= 0x10;
     orPort(LCD_E_PORT, LCD_E_MASK);   // Enable  set 1;
     asm volatile ("nop");  // internal delay
     asm volatile ("nop");  // data setup
     in1 = inPort(LCD_D4_PORT);  // in D4 port nib bit 0
     in2 = brickPervertAlignR(inPort(LCD_DATA_PORT));
     addr |= in2;
     andPort(LCD_E_PORT, ~LCD_E_MASK); // Enable  set 0;
     if (in1 & LCD_D4_MASK) addr |= 0x01;
#elif LCD_DATA_USE == ALL_BITS
     orPort(LCD_E_PORT, LCD_E_MASK);   // Enable  set 1;
     asm volatile ("nop");  // internal delay
     asm volatile ("nop");  // data setup
     addr = inPort(LCD_DATA_PORT);
     andPort(LCD_E_PORT, ~LCD_E_MASK); // Enable  set 0;
    // delay needed ??
#elif LCD_DATA_USE == LOWER_NIBBLE
    // as no fitting hardware target platform is around this part
    // (LCD module high nibble on port low nible) is not tested yet.
     orPort(LCD_E_PORT, LCD_E_MASK);   // Enable  set 1;
     asm volatile ("nop");  // internal delay
     asm volatile ("nop");  // data setup
     uint8_t in1 = inPort(LCD_DATA_PORT); // high nibble first
     andPort(LCD_E_PORT, ~LCD_E_MASK); // Enable  set 0;
     addr = (in1 << 4) & 0xF0;
     orPort(LCD_E_PORT, LCD_E_MASK);   // Enable  set 1;
     asm volatile ("nop");  // internal delay
     asm volatile ("nop");  // data setup
     in1 = inPort(LCD_DATA_PORT) & 0xF0; // low nibble second
     andPort(LCD_E_PORT, ~LCD_E_MASK); // Enable  set 0;
     addr |= in1;
 #elif LCD_DATA_USE == UPPER_NIBBLE
    // upper nibble first
     orPort(LCD_E_PORT, LCD_E_MASK);   // Enable  set 1;
     asm volatile ("nop");  // internal delay
     asm volatile ("nop");  // data setup
     uint8_t in1 = inPort(LCD_DATA_PORT); // high nibble first
     andPort(LCD_E_PORT, ~LCD_E_MASK); // Enable  set 0;
     addr = in1 & 0xF0;
     orPort(LCD_E_PORT, LCD_E_MASK);   // Enable  set 1;
     asm volatile ("nop");  // internal delay
     asm volatile ("nop");  // data setup
     in1 = inPort(LCD_DATA_PORT); // low nibble second
     andPort(LCD_E_PORT, ~LCD_E_MASK); // Enable  set 0;
     addr |= (in1 >> 4) & 0x0F;
#else
#error read flag not implemented
#endif

  orPort(LCD_RS_PORT, LCD_RS_MASK); // RS  set 1
  setLCDdataPorts4out();
  return addr;
#else
  return 0xFF;
#endif
} // addressFromLCD()


/* Wait for busy flag gone with short (50s) or long (1.6 ms) timeout
 *
 *  This function is not available if the R/W on the TC1602B is
 *  \ref  RW_NOTFIXED "fixed" to Gnd.
 *
 *  This function uses spin waiting for reading the address and busy flag
 *  until the busy flag is gone or a timeout. For the background see
 *  \ref pause4LCDmodule.
 *
 *  The return value is the last read. If bit 7, the busy flag, is set, this
 *  function returned because of timeout.
 *  Hint internal: addressFromLCD() takes min 1.8s on 4-bit mode. One while
 *  loop step is 3.5s.
 *
 *  \param loPa true: >= 1.6 ms  timeot; false: >= 50 s timeout
 */
uint8_t waitWhileLCDbusy(uint8_t  loPa){
#if  RW_NOTFIXED
   uint16_t i = loPa ? 520 : 15; // this is OK for 16 MHz
   while (1) {
      if (! (--i)) break;
      loPa =  addressFromLCD();
      if (! (loPa &0x80)) break;
   }
#else
   pause4LCDmodule(loPa);
#endif
   return loPa;
}  // waitWhileLCDbusy(uint8_t)


/*  Short (37s) or long (1.52 ms) pause
 *
 *  This is just an initialisation helper function.
 *  The initialisation procedere of the TC1602B LCD controller requires two
 *  waits of at least 37s duration and one wait of at least 1520s. This
 *  quite obscure requirement is independent of 4 or 8 bit mode.
 *  See the respective manufacturer's data sheet.
 *
 *  In simple cases this might be implemented by spin waiting.
 *  In real time automation environments, as weAutSys, this should be done by
 *  the yielding and timer functions featured by the run time.
 *
 *  This implementation is (stupid) spin waiting based on an existing (serial
 *  bootloader) function already in upper flash.
 *
 *  \param loPa true: the long pause
 */
void pause4LCDmodule(uint8_t const loPa){
   uint8_t i = loPa ? 1520 / 25 + 2 : 2;
   for (;i; --i) {  wait25(); }
} // pause4LCDmodule(uint8_t

void pause4LCDinit(uint8_t const loPa){
   uint8_t i = loPa ? 1520 / 25 + 2 : 2;
   for (;i; --i) {  wait25(); }
} // pause4LCDinit(uint8_t)

/*  A pause of  n * 100ms
 *
 *  This is a pure courtesy helper function for display timing.
 *  It is implemented here by just spin waiting. All the warnings and
 *  hints given with preInitLCDpause() apply.
 *
 *  \param mult: 1..255: 0.1s .. 25.5s pause
 */
void pause100ms(uint8_t mult){
   while (mult) {
      uint16_t cnt = 100000 / 25;
      do { wait25(); } while (--cnt);
      --mult;
   }
} // pause100ms(uint8_t)


//---  LCD module initialisation                ----------------------------

/*  Initialise the LCD
 *
 *  This function initialises the LCD controller completely. After power up
 *  this is mandatory.
 *
 *  Some operation modes may be chosen here:
 *  N true: two line mode instead of one line
 *  F true: 5*11 matrix for characters instead of 5*8
 *  C true: cursor visible
 *  B true: the display is blinking
 *
 *  The usual standard parameter value is 0x08: two lines of 16 5*8 characters
 *  each, no cursor, steady display.
 *
 *  \param modeNFCB bits 3210 set is the respective NFCB true
 */
void initLCD(uint8_t modeNFCB){
   setLCDcontrolPorts();  // set the port bits used as LCD interface
   send1st4bitModeNibble(3);   // reset to 8 bit no matter what
   pause4LCDinit(0);
   send1st4bitModeNibble(3);  // set to 8 bit mode (ignored  if was 4 bit)
   pause4LCDinit(0);
#if LCD_DATA_USE == ALL_BITS
#else // 8 bit else 4 bit mode
   send1st4bitModeNibble(2); // set to 4 bit mode from guaranteed 8 bit mode
   pause4LCDinit(0);         // with lower nibble wrong
// init problems try remedy
   sendByte2LCD(0x28); // repeat standard init (
   pause4LCDinit(0);
#endif
   sendByte2LCD(0x20 | (0x0C & modeNFCB)); // function set with NF  (28)
   pause4LCDinit(0);
   sendByte2LCD(0x0C | (0x03 & modeNFCB)); // display control with CB
   pause4LCDinit(0);
   // from now on BF (busy flag) can be used if R/W isn't soldered to Gnd
   sendByte2LCD(0x01); // clear display
   waitWhileLCDbusy(1);
   sendByte2LCD(0x06); // entry set mode increment on, shift off
   waitWhileLCDbusy(0);
} // initLCD()

/*  Set ports used for LCD control as output.
 *
 *  This is a pure intialisation helper function.
 */
void setLCDcontrolPorts(){
#if LCD_DATA_USE != 0
#if RW_NOTFIXED
   andPort(LCD_RW_PORT, ~LCD_RW_MASK); // RW  set 0
   outPins(LCD_RW_PORT, LCD_RW_MASK); //RW set out
#endif
   andPort(LCD_DATA_PORT, ~LCD_DATA_USE); // 3,4,8 bits for data set 0
#if LCD_DATA_USE == BRICK_PERVERT
   andPort(LCD_D4_PORT, ~LCD_D4_MASK); // extra (pervert) LCD-D  set 0
#endif
   setLCDdataPorts4out();
   orPort(LCD_RS_PORT, LCD_RS_MASK); // RS  set 1 (risk stale data not com.)
   outPins(LCD_RS_PORT, LCD_RS_MASK); // RS set out
   andPort(LCD_E_PORT, ~LCD_E_MASK); // E  set 0
   outPins(LCD_E_PORT, LCD_E_MASK); //E set out   may / will latch stale data

   /// { problem of single init not solved
   orPort(LCD_E_PORT, LCD_E_MASK); // E  set 1 for 2. stale if in 4 bit mode
   andPort(LCD_E_PORT, ~LCD_E_MASK); // E  set 0 may / will latch stale lo
   /// not solved  may be removed }

   andPort(LCD_RS_PORT, ~LCD_RS_MASK); // RS  set 0 (command mode for init)
#endif
} // setLCDcontrolPorts()

/** Set ports used for LCD data as output.
 *
 *  This is a pure helper function for initialisation and if feasible
 *  for switching from (data) input back to output.
 *  This function also sets the R/W control line (if given) to 0.
 */
void setLCDdataPorts4out(){
#if LCD_DATA_USE != 0
#if RW_NOTFIXED
   andPort(LCD_RW_PORT, ~LCD_RW_MASK); // RW  set 0
#endif
   outPins(LCD_DATA_PORT, LCD_DATA_USE); //bits for data set out
#ifdef  LCD_D4_PORT
   outPins(LCD_D4_PORT, LCD_D4_MASK); // extra (pervert) LCD-D  set 0
#endif
#endif
} // setLCDdataPorts4out()

/** Set ports used for LCD data as input.
 *
 *  This is a pure helper function for switching from (data) output to
 *  input. If the R/W line is \ref RW_NOTFIXED "fixed" to ground this function
 *  does nothing. <br />
 *  Otherwise this function also sets the R/W control line to 1.
 */
void setLCDdataPorts4in(){
#if LCD_DATA_USE != 0
# if RW_NOTFIXED
#  ifdef  LCD_D4_PORT
  inpPins(LCD_D4_PORT, LCD_D4_MASK);
#  endif
  inpPins(LCD_DATA_PORT, LCD_DATA_USE); // data port in
  orPort(LCD_RW_PORT, LCD_RW_MASK);  // RW on
# endif
#endif
} // setLCDdataPorts4in()


//---  LCD module user defined character pattern initialisation   ----------

/*  Set a pattern in the character generator RAM
 *
 * \param code target character code 0 .. 7
 * \pattern address of a an array of 8 bytes, one per line, in flash memory
 */
void setCharPatternLCD(uint8_t code, ADDR_T pattern){
#if LCD_DATA_USE != 0
  sendCommand2LCD(0x40 | ((code & 0x07) *8));
  orPort(LCD_RS_PORT, LCD_RS_MASK); // RS  set 1
  pause4LCDmodule(0);  // 39 s
  for (uint8_t i=0; i < 8; ++i) {
    uint8_t c = getFlashByte(pattern++);
    sendByte2LCD(c);
    waitWhileLCDbusy(0);  // 43 s
  } // for over the 8 lines
#endif
} // setCharPatternLCD(uint8_t, ADDR_T)


/** A DIN resistor symbol (upright) as user defined character pattern. */
INFLASH(uint8_t const resistor[]) = {
 ____X____
 ____X____
 __X_X_X__
 __X___X__
 __X___X__
 __X___X__
 __X_X_X__
 ____X____
};

/** A DIN inductor symbol (upright) as user defined character pattern. */
INFLASH(uint8_t const inductor[]) = {
 ____X____
 ____X____
 __X_X_X__
 __X_X_X__
 __X_X_X__
 __X_X_X__
 ____X____
 ____X____
};

/** A capacitor symbol (upright) as user defined character pattern. */
INFLASH(uint8_t const capacitor[]) = {
 ____X____
 ____X____
 X_X_X_X_X
 _________
 X_X_X_X_X
 ____X____
 ____X____
 ____X____
};

/** A diode symbol (downward) as user defined character pattern. */
INFLASH(uint8_t const diode[]) = {
 ____X____
 ____X____
 X_X_X_X_X
 __X___X__
 ____X____
 X_X_X_X_X
 ____X____
 ____X____
};

/** A diode symbol (upward) as user defined character pattern. */
INFLASH(uint8_t const diodeUp[]) = {
 ____X____
 ____X____
 X_X_X_X_X
 ____X____
 __X___X__
 X_X_X_X_X
 ____X____
 ____X____
};

/** A electrolytic capacitor symbol (+ upward) as character pattern. */
INFLASH(uint8_t const elCo[]) = {
 ____X____
 ____X____
 X_X_X_X_X
 X_______X
 X_X_X_X_X
 _________
 X_X_X_X_X
 ____X____
};


//---  commands to the LCD module               ----------------------------

/*  Clear the  LCD
 *
 *  This function makes all character display blank (empty), but does
 *  not influence backlight, brightness or else.
 */
void clearLCD(){
#if LCD_DATA_USE != 0
   sendCommand2LCD(1);
   pause4LCDmodule(1);
#endif
} // clearLCD()

/*  Home LCD cursor
 *
 *  This function sets the cursor (next) character write to the leftmost
 *  position in the upper line. cursPosLCD(1, 0) is usually faster.
 */
void cursHomLCD(){
#if LCD_DATA_USE != 0
   sendCommand2LCD(2);  // try 0x80 !
   pause4LCDmodule(1);
#endif
} // cursHomLCD()

/*  Set cursor position
 *
 *  This function sets the cursor respectively next character write position
 *  to the line and col position.
 *  A TC1602B LCD module has two lines (1 &amp; 2)
 *  of 16 columns (0..15) each.
 *
 *  \param line 0 or 1: upper line; 2 (or else): lower line
 *  \param col  0 ..15: left to right; is taken modulo 16.
 */
void cursPosLCD(uint8_t line, uint8_t col){
#if LCD_DATA_USE != 0
   sendCommand2LCD((line > 1 ? 0xC0 : 0x80) | (col & 0x0F));
   waitWhileLCDbusy(0);
   // pause4LCDmodule(0); // wait25();
#endif
} // lowerLineLCD()


//---  data / text to the LCD module            ----------------------------


/*  Send data (byte) to LCD  */
void sendData2LCD(uint8_t data){
#if LCD_DATA_USE != 0
# if RW_NOTFIXED
   andPort(LCD_RW_PORT, ~LCD_RW_MASK); // RW  set 0
# endif
   orPort(LCD_RS_PORT, LCD_RS_MASK); // RS  set 1
   sendByte2LCD(data);
   waitWhileLCDbusy(0);
#endif
} // sendData2LCD(uint8_t)


/*  Send raw data string (multiple bytes) to LCD module from RAM
 *
 *  This is a very basic send function for the initialised LCD module. It
 *  is just based on \ref sendData2LCD and uses spin waiting for the required
 *  inter-character pause. Hence all warnings regarding real time
 *  applications apply.
 *
 *  Raw means that every byte including 0 is sent as is.
 *
 *  \param src pointer to the first character or byte to be sent to the
 *             LCD module. This is not a 0-terminated string. This must
 *             must not be null -- but is at present tolerates in this
 *             implementation
 *  \param n  the number of bytes to sent; n must be in the range 0..16,
 *            but never more then space remaining on the actual line.
 *            <=16 is checked in this implementation.
 */
void sendRawString2LCD(char const * src, uint8_t n){
#if LCD_DATA_USE != 0
# if RW_NOTFIXED
  andPort(LCD_RW_PORT, ~LCD_RW_MASK); // RW  set 0
# endif
  orPort(LCD_RS_PORT, LCD_RS_MASK); // RS  set 1
  if (n > 16) return; // wrong number
  if (!src) return; // no string
  for(;n;--n) { // for over string
     uint8_t c = *src++;
     sendByte2LCD(c);
     waitWhileLCDbusy(0);
  }  // for over string
#endif
} // sendRawString2LCD(char const *, uint8_t)

/** Send raw data string (multiple bytes) to LCD module from flash memory
 *
 *  This is the same function as \ref sendRawString2LCD, except that the raw
 *  string or character array is expected in program (flash) memory.
 */
void sendRawString2LCD_P(ADDR_T src, uint8_t n){
#if LCD_DATA_USE != 0
# if RW_NOTFIXED
  andPort(LCD_RW_PORT, ~LCD_RW_MASK); // RW  set 0
# endif
  orPort(LCD_RS_PORT, LCD_RS_MASK); // RS  set 1
  if (n > 16) return; // wrong number
  if (!src) return; // no string
  for(;n;--n) {
     uint8_t c = getFlashByte(src++);
     sendByte2LCD(c);
     waitWhileLCDbusy(0);
  }
#endif
} // sendRawString2LCD(char const *, uint8_t)


//---  helper functions for the LCD module   (internal use only)    --------

/*  Send 8 bit (full byte) to LCD
 *
 *  Pure helper function, usually not called by application software.
 */
  void sendByte2LCD(uint8_t const data){
#if LCD_DATA_USE != 0
    orPort(LCD_E_PORT, LCD_E_MASK);   // Enable  set 1;
# if LCD_DATA_USE == BRICK_PERVERT
    send1st4bitModeNibble(data >> 4); // high nibble first then the
    send1st4bitModeNibble(data); // low one using the pervert brick's helper
# elif LCD_DATA_USE == ALL_BITS
    toPort(LCD_DATA_PORT, data);
    // delay needed ??
# elif LCD_DATA_USE == LOWER_NIBBLE
    // as no fitting hardware target platform is around this part
    // (LCD module high nibble on port low nible) is not tested yet.

    // upper nibble first
    uint8_t dts = (data >> 4) & 0x0F;  // GGC uses swap OK
    andPort(LCD_DATA_PORT, ~LCD_DATA_USE); // set 0
    orPort(LCD_DATA_PORT, dts); // or upper data nibble     ---+
    dts = data & 0xF0;   //                                    |
    andPort(LCD_E_PORT, ~LCD_E_MASK); // Enable  set 0   latch /
    orPort(LCD_E_PORT, LCD_E_MASK);   // Enable  set 1;
    andPort(LCD_DATA_PORT, ~LCD_DATA_USE); // set 0
    orPort(LCD_DATA_PORT, dts);  // or lower data nibble   --->+

# elif LCD_DATA_USE == UPPER_NIBBLE
    // upper nibble first
    uint8_t dts = data & 0xF0;
    andPort(LCD_DATA_PORT, ~LCD_DATA_USE); // set 0
    orPort(LCD_DATA_PORT, dts); // or upper data nibble     ---+
    dts = (data << 4) & 0xF0;   // GGC uses swap OK            |
    andPort(LCD_E_PORT, ~LCD_E_MASK); // Enable  set 0   latch /
    orPort(LCD_E_PORT, LCD_E_MASK);   // Enable  set 1;
    andPort(LCD_DATA_PORT, ~LCD_DATA_USE); // set 0
    orPort(LCD_DATA_PORT, dts);  // or lower data nibble   --->+
# endif                    //                                   |
    andPort(LCD_E_PORT, ~LCD_E_MASK); // Enable  set 0   latch /
#endif
  } // sendByte2LCD(uint8_t)


/* Send first 4 bit (mode nibble) to LCD
 *
 *  Pure helper function, not to be called by application software and
 *  never for an 8 bit transfer configuration.
 *
 *  \param nib 1..0xF
 */
void send1st4bitModeNibble(uint8_t nib){
#if LCD_DATA_USE != 0
   orPort(LCD_E_PORT, LCD_E_MASK);   // Enable  set 1;
# if LCD_DATA_USE == BRICK_PERVERT
   andPort(LCD_DATA_PORT, ~LCD_DATA_USE); // set data port 0 for D0..3
   orPort(LCD_DATA_PORT, brickPervertAlignW(nib)); //set data port acc. to nib
   if (nib & 1) { // bit 4
      orPort(LCD_D4_PORT, LCD_D4_MASK);  // D4 set 1
   } else {
      andPort(LCD_D4_PORT, ~LCD_D4_MASK); // D4 set 0
   }
   //   for (uint8_t i = 3; --i;) {asm volatile("nop "); }
# elif LCD_DATA_USE == ALL_BITS
 // this function must not be called on 8 bit transfer LCD
# elif LCD_DATA_USE == LOWER_NIBBLE
   andPort(LCD_DATA_PORT, ~LCD_DATA_USE); // set 0
   orPort(LCD_DATA_PORT, (nib & LCD_DATA_USE)); // to lower port nibble
# elif LCD_DATA_USE == UPPER_NIBBLE
    andPort(LCD_DATA_PORT, ~LCD_DATA_USE); // set 0
    orPort(LCD_DATA_PORT, ((nib << 4) & LCD_DATA_USE)); // to upper data nibble
# endif
   andPort(LCD_E_PORT, ~LCD_E_MASK); // Enable  set 0;
#endif
} // send1st4bitModeNibble()


/*  Send command (byte) to LCD
 *
 *  This is a helper function, normally not used by application
 *  software. <br />
 *  It sets the command (write) mode and sends the data byte to the LCD
 *  module. It does not include the required command specific pause.
 *
 *  \param command byte (see TC1602B data sheet)
 */
void sendCommand2LCD(uint8_t data){
#if LCD_DATA_USE != 0
# if RW_NOTFIXED
   andPort(LCD_RW_PORT, ~LCD_RW_MASK); // RW  set 0
# endif
   andPort(LCD_RS_PORT, ~LCD_RS_MASK); // RS  set 0
   sendByte2LCD(data);
#endif
} // sendCommand2LCD(uint8_t)
