/**
 * \file tickTime.c
 *
 * Implementation of some timing and tick interrupt functions
 *
 *  Copyright (c) 2014, 2015  Albrecht Weinert, Bochum
 *
 * This library is just a copy (a very small subset to be precise) of the
 * timing functions in weAutSys. <br />
 * See all else in the weAutSys documentation. <br />
 *  To do: Try to replace this copy and paste by a better approach in future.
 *
 * $Revision: 1 $
 * $Date: 2017-01-25 11:27:06 +0000 (Wed, 25 Jan 2017) $
 */

#include "tickTime.h"
// #include <avr/io.h> // comment in just to please Eclipse, if all else fails
// #include <avr/iom32.h>
// #include <avr/iom2560.h>
// #include <avr/iom2561.h>
// #include <avr/iom1284p.h>

// next three lines just to please Eclipse
#ifndef USE_TESTPINS
# define USE_TESTPINS 1
#endif

// ------------------ tick interrupt ------------------------------------

uint8_t volatile tick__1_ms;
uint8_t volatile cn12u8;
uint8_t volatile tick100_ms;

uint8_t vcoCount  = VCO_DEFAULT; //!< value controlled oscillator (NTP)
uint8_t vcoSetVal = VCO_DEFAULT; //!< holds best fit for 20 MHz Quartz

/* Trimming of timing oscillator
 *
 *  This function will set the "value controlled oscillator" for the
 *  system's 1 ms tick to normal (usually best) speed.
 */
void setVCOnormal(void){  vcoSetVal = VCO_DEFAULT; }
uint8_t getVCOsetting(void){ return  vcoSetVal; }


/* Trimming of timing oscillator
 *
 *  This function will speed up the "value controlled oscillator" for the
 *  system's 1 ms tick a little bit. This is for catching up quite little
 *  differences to an externally given absolute (NTP e.g.) time.
 *
 *  Little differences in that sense are below -200 ms.
 */
void speedVCOup(void){ if (vcoSetVal) ++vcoSetVal; }

/*  Trimming of timing oscillator
 *
 *  This function will slow down the "value controlled oscillator" for the
 *  system's 1 ms tick a little bit. This is for catching up quite little
 *  differences to an externally given absolute (NTP e.g.) time.
 *
 *  Little differences in that sense are below +100 ms.
 */
void slowVCOdown(void){ if (! --vcoSetVal) vcoSetVal = 1; }

uint8_t next_100;

/** Tick timer interrupt handler
 *
 *  TIMER0 is \ref intro_sec "weAutSys"' "playground" only. User software
 *  must never ever interfere with it.
 */
ISR(MS_TICK_vect){
   if ((++tick__1_ms) == next_100) { // 100ms tick
      next_100 +=100;
      ++tick100_ms;
#if (USE_TESTPINS == 1) && 1
      setTestPin1(tick100_ms & 1); // TEST of 100 ms tick
#endif
   } // 100ms tick
#if (USE_TESTPINS == 1)
   setTestPin0(tick__1_ms & 1); // TEST of 1 ms tick
#endif

   cn12u8 += MS_TICK_OCR + 1; // advance 16s (16MHz) / 12.5s (20MHz) counter base
   if (! --vcoCount) {  // corriger 0,16 % too fast to 1,12% too slow
#if USE_TESTPINS == 1 & 0
      setTestPin1(ON); // TEST VCO
#endif
      vcoCount = vcoSetVal;
      MS_TICK_OCR  = PRESC_TRIM_FAC;     // count one more
   }  else {   // let it be:  0,018 % too fast
#if USE_TESTPINS == 1  & 0
      setTestPin1(OFF); // Test VCO
#endif
#if defined(EXTRA_SLOW)
      if (vcoCount <= EXTRA_SLOW) {
         MS_TICK_OCR  = PRESC_TRIM_FAC;     // count one more
      } else  // no opening brace: else is the one statement below
#endif
       MS_TICK_OCR  = PRESC_TRIM_FAC - 1; // count normally
   }
} // timer 0 match interrupt


/*  Get the 12.8 (16) s tick value (8 bit) */
uint8_t cnt12u8_8(void){ return cn12u8 + TCNT0; }


/* low level clock / timer / tick initialisation */
void initSystTiming(void){
   uint8_t const sreg_save = SREG; cli(); // disable interrupt
   MS_TICK_OCR  = PRESC_TRIM_FAC - 1;
#if defined(TIMER0_COMPA_vect)
   TCCR0A = 0x02; // no output;     WGmode=2 (Bits01)
# if  PRESC_SET == 64
   TCCR0B = 0x03; // prescaler = 64;
# else
   TCCR0B = 0x04; // prescaler = 256;
# endif
   OCR0B  = 0xFF; // compare B not used
   TIMSK0 = 0x02; // compare A interrupt enable
#elif defined(TIMER0_COMP_vect)
   //  todo init small C
# if  PRESC_SET == 64
   TCCR0 = 0x0B; // prescaler = 64; WGmode=2 (Bit2) no output
# else
   TCCR0 = 0x0C; // prescaler = 256; WGmode=2 (Bit2) no output
# endif
   TIMSK = 0x02; // compare timer0 interrupt enable
#endif
   cn12u8 = tick__1_ms = tick100_ms = 0;
   next_100 = 100;
   SREG = sreg_save; // restore interrupt
} // initSystTiming()


/*  Initialisation of the (potential) test pins
 *  \param TP0start initial value for Testpin 0 (ON or OFF)
 *  \param TP1start initial value for Testpin 1 (ON or OFF)
 */
void initTestPins(uint8_t TP0start, uint8_t TP1start){
#ifdef TEST_PIN0_PORT
   setTestPin0(TP0start); // reg and pullup
   outPins(TEST_PIN0_PORT, (1<<TEST_PIN0_PIN));
#endif
#ifdef TEST_PIN1_PORT
   setTestPin1(TP1start);
   outPins(TEST_PIN1_PORT, (1<<TEST_PIN1_PIN));
#endif
} // initTestPins()


void setTestPin0(uint8_t state){
#if !defined(_NO_TESTPINS_) && defined(TEST_PIN0_PORT)
   if (state) {
      orPort(TEST_PIN0_PORT, (1 << TEST_PIN0_PIN));
   } else {
      andPort(TEST_PIN0_PORT, ~(1 << TEST_PIN0_PIN));
   }
#endif
} // setTestPin0(uint8_t)

void setTestPin1(uint8_t state){
#if !defined(_NO_TESTPINS_) && defined(TEST_PIN1_PORT)
   if (state) {
      orPort(TEST_PIN1_PORT, (1 << TEST_PIN1_PIN));
   } else {
      andPort(TEST_PIN1_PORT, ~(1 << TEST_PIN1_PIN));
   }
#endif
} // setTestPin1(uint8_t)





