/** \file lcd1602b.h
 *
 *  Definitions of LCD module (1602B types) functions
 *
 *  Copyright (c) 2015  Albrecht Weinert, Bochum
 *
 *  Currently we support three hardware target platforms (PLATFORM=..): <br />
  - easyAVR32
  - arduMega2560
  - boAVR2013
  - arduinoUno
 *
 *  The implementation is found in file extra/lcd1602b.c. <br />
 *  lahn.c is a little test application for above named target
 *  platforms. <br />
 *  The properties of target platforms supported so far are described
 *  within platform-specific include files named arch/expConf_PLATFORM.h
 *  in a special section "platform specifica for LCD module TC1602B".
 *  By this scheme further platforms and variants are easily added conforming
 *  to the scheme found there.
 *
 *  1
 *  25.01.2017
 */
#ifndef LCD1602B_H
#define LCD1602B_H

#include <stdint.h>
#include "arch/expConf.h"
#include "we-aut_sys/ll_common.h"  // just to please Eclipse

/** LCD data interface mode
 *
 *  Use only 4 bits: upper nibble of port \ref LCD_DATA_PORT.
 *
 *  \showinitializer
 */
#define UPPER_NIBBLE 0xF0

/** LCD data interface mode
 *
 *  Use only 4 bits: lower nibble of  port \ref LCD_DATA_PORT.
 *
 *  \showinitializer
 */
#define LOWER_NIBBLE 0x0F

/** LCD data interface mode
 *
 *  Use all 8 bits: take port \ref LCD_DATA_PORT as one byte.
 *
 *  \showinitializer
 */
#define ALL_BITS 0xFF

/** LCD data interface mode
 *
 *  Use all 4 bits: take 3 bits D5..D7 from bits 0 to 3 of port
 *  \ref LCD_DATA_PORT. <br />
 *  Note: The upper three bits of the nibble need a
 *  \ref brickPervertAlignW "shift for the write" to and another
 *  \ref brickPervertAlignW "shift after the read" from the port
 *  \ref LCD_DATA_PORT. <br />
 *  Take LCD D4 respectively nibble bit 0 as bit masked by \ref LCD_D4_MASK
 *  of the \ref LCD_RW_PORT.
 *
 *  This setting is for the "Electronic Brick Starter Kit" for Arduino to
 *  be plugged on a "ArduinoMega 2560" (make PLATFORM=arduMega2560). <br />
 *  Really: It is hardly believable, but in this (commercially available)
 *  hardware configuration the LCD module's 4 data bits used will be connected
 *  to two different ATmega ports. Additionally this is done in a quite
 *  arbitrary way; refer to the macro \ref brickPervertAlignW.
 *  \showinitializer
 */
#define BRICK_PERVERT 0x07

#define brickPervertAlignW(d3) (((d3) >> 1) & BRICK_PERVERT)
#define brickPervertAlignR(d3) (((d3) & BRICK_PERVERT) << 1)


// some platform make tips (just notes / memo)

// for boAVR2013
// make PLATFORM=boAVR2013  MAIN_F=lahn PRG_APP_P=com5  clean  all
// make PLATFORM=boAVR2013  MAIN_F=lahn PRG_APP_P=com5  progapp

// foreasyAVR32
// make PLATFORM=easyAVR32  MAIN_F=lahn PRG_APP_P=com14  clean all
// make PLATFORM=easyAVR32  MAIN_F=lahn PRG_APP_P=com14  progapp

// for arduinoUno  (and eBrickOnArduinoUno)
// make PLATFORM=arduinoUno  MAIN_F=lahn PRG_APP_P=com11  clean  all
// make PLATFORM=arduinoUno  MAIN_F=lahn PRG_APP_P=com11  progapp


// for arduMega2560
// make PLATFORM=arduMega2560  MAIN_F=lahn PRG_APP_P=com5  clean  all
// make PLATFORM=arduMega2560  MAIN_F=lahn PRG_APP_P=com5  progapp


#ifndef LCD_DATA_USE
// do not include this file lcd1602b.h on platforms withot this LCD module
// or with no possibility to add one.
// If this cannot be avoided (as it should), make next line a warning
#  error "no LCD display for" PLATFORM
#endif

#ifndef RW_NOTFIXED  // in most cases the R/W line is fixed to 0 (write)
# define RW_NOTFIXED 0
#endif

#ifndef SUPPLY_VOLTAGE // TC1602B's timing partly depends on supply voltage
# define SUPPLY_VOLTAGE 5
#endif

//--  application functions for LCD module (TC1602B) usage      -------------

/** Initialise the LCD
 *
 *  This function initialises the LCD controller completely. After
 *  power up this is mandatory.
 *
 *  Some operation modes may be chosen here: <br />
 *  N true: two line mode instead of one line <br />
 *  F true: 5*11 matrix for characters instead of 5*8 <br />
 *  C true: cursor visible <br />
 *  B true: the display is blinking <br />
 *
 *  The usual standard is 0x08: two lines of 16 5*8 characters each, no cursor,
 *  steady display.
 *
 *  \param modeNFCB bits 3210 set means setting the respective NFCB true
 */
void initLCD(uint8_t modeNFCB);

 /** Send data to the LCD module
  *
  *  This function sends a byte to the LCD module. If in proper mode and state
  *  the character denoted by the code \c data would afterwards be displayed
  *  at the current (cursor / write) position, that would be (post-)
  *  incremented to the next column. <br />
  *  This function includes the necessary \ref pause4LCDmodule "pause"
  *  after that write / display  operation.
  *
  \param data byte send to the LCD module (normally interpreted as character)
  */
 void sendData2LCD(uint8_t data);

 /** Send raw data string (multiple bytes) to the LCD module from RAM
  *
  *  This is a very basic send function for the initialised LCD module. It
  *  is just based on \ref sendData2LCD and uses spin waiting for the required
  *  inter-character pause. Hence all warnings regarding real time
  *  applications apply. (In short: Do not use this function there.)
  *
  *  Raw means that every byte including 0 is sent as is. It will be displayed
  *  as decided by the respective module. <br />
  *  Note: Except for the range 32..126 (pure US ASCII) the character
  *  displayed totally depends of the special TC1602B make in hand and partly
  *  on user programmed character generators. <br />
  *  Note 2: The cheapest modules feature simplified Chinese. That renders
  *  about half the character patterns useless for European contexts.
  *
  *  \param src pointer to the first character or byte to be sent to the
  *             LCD module. This is not a 0-terminated string.
  *  \param n   the number of bytes to sent; n must be in the range 0..16,
  *             but never more then space remaining on the actual line
  *             (caller's responsibility)
  *  \sa sendRawString2LCD_P
  */
 void sendRawString2LCD(char const * src, uint8_t n);

 /** Send raw data string (multiple bytes) to LCD module from flash memory
  *
  *  This is the same function as \ref sendRawString2LCD, except the raw
  *  string or character array being expected in program (flash) memory.
  */
 void sendRawString2LCD_P(ADDR_T src, uint8_t n);

 /** Clear the  LCD
  *
  *  This function makes all character display blank (empty), but does
  *  not influence backlight, brightness or else.
  */
 void clearLCD(void);

 /** Home LCD cursor
  *
  *  This function sets the cursor (or next character write) to the leftmost
  *  position in the upper line and does some shifting if mode so set. <br />
  *  \ref curPosLCD "cursPosLCD(0, 0)" is faster.
  */
 void cursHomLCD(void);

 /** Set cursor position
  *
  *  This function sets the cursor respectively the next character write
  *  position to the \c line and \c col position. <br />
  *  A TC1602B LCD module has two lines (1 &amp; 2)
  *  of 16 columns (0..15) each.
  *
  *  \param line 0 or 1: upper line; 2 (or else): lower line
  *  \param col  0 ..15: left to right; is taken modulo 16.
  */
 void cursPosLCD(uint8_t line, uint8_t col);


/** A pre-initialisation pause of 20ms or 30 ms
 *
 *  The LCD modules of TC1602B type need an astonishingly long pause from
 *  power up to the \ref initLCD "initialisation".
 *
 *  This function should be called before \ref initLCD if the LCD module is
 *  to be initialised immediately after power up. <br />
 *  See all else in \ref pause4LCDmodule.
 *
 *  This function will be provided for simple cases or demo applications only.
 *  It is implemented by spin waiting! In real time automation environments,
 *  as \ref intro_sec "weAutSys", this absolutely must be done by by using the
 *  \ref pt "yielding" and \ref timerHdl "timer" functions featured by the
 *  \ref intro_sec "run time". Never use this function (provided as courtesy
 *  for simple cases) in a professional application.
 *
 *  \param sup3V 3: the supply is only 3V: the longer wait is needed
 */
void preInitLCDpause(uint8_t sup3V);

/** A pause of 100ms of multiples
 *
 *  This is a pure courtesy helper function for display timing. <br />
 *  It might here be implemented by spin waiting. All the warnings and
 *  hints given with \ref preInitLCDpause apply. <br />
 *  See all else in \ref pause4LCDmodule.
 *
 *  \param mult: 1..255: 0.1s .. 25.5s pause
 */
void pause100ms(uint8_t mult);

/** Set pattern to character generator RAM
 *
 *  This function sets the 5*8 matrix pattern of one user defined character.
 *  The TC1602 features 8 user defined characters, which can be defined and
 *  displayed under \c code 0..7. They may also be displayed by the (raw)
 *  code 8..15 *1) as well. <br />
 *  Note 1: Using the codes 0..15 for displayable characters makes the
 *  definition of a string output function implementing the the expected
 *  behaviour of control codes like \ref LF, \ref RET, FF etc. as well as of
 *  0 for C's string termination hardly feasible.
 *
 *  The user defined \c pattern is given as (address of -- see note *2)) an
 *  array in flash memory of 8 byte, [0] being the upper line and [7] the
 *  lowest line of the 5*8 matrix. <br />
 *  Note 2: To get the flash address of the pattern array in question, use one
 *  of the macros \ref LOW_ADD or (if in high bootloader memory) FAR_ADD.
 *
 *  In each line of the pattern or each byte of the flash pattern array only
 *  the lower five bits (for the matrix's columns) are relevant. To define
 *  the whole pattern in human readable form use the 32 macros
 *  \ref   _________ ... \ref X_X_X_X_X  provided here. <br />
 *  Note 3: Keep in mind those macros each containing the comma needed as
 *  separator in array initialisers. So the pattern defined is as unspoiled
 *  for the human as feasible. Use your imagination to see a X as lit and an
 *  _ as off.
 *
 *  Call this function by e.g.: \code
 *   setCharPatternLCD(0, LOW_ADD(patternDiode)); \endcode
 *
 *  Define the pattern used by e.g.: \code
 *   INFLASH(uint8_t const diode[]) = {
 *     ____X____
 *     ____X____
 *     X_X_X_X_X
 *     __X___X__
 *     ____X____
 *     X_X_X_X_X
 *     ____X____
 *     ____X____
 *  };  \endcode
 *
 *  Note 3: To resume normal character display operation after defining the
 *  last pattern by this function one of the functions \ref cursPosLCD,
 *  \ref clearLCD or \ref cursHomLCD must be called.
 *
 *  \param code target character code 0 .. 7
 *  \pattern address of a an array of 8 bytes, one per line, in flash memory
 */
void setCharPatternLCD(uint8_t code, ADDR_T pattern);


// Human readable patterns to generate an array of 8 bytes -- one per line --
// in flash memory for loadable character pattern.
//
// in   X_X_X_X_X  the marked positions only X is dot on and _ is off.
#define _________ 0x00 , //!< readable pattern for 1602 user defined characters
#define ________X 0x01 , //!< readable pattern for 1602 user defined characters
#define ______X__ 0x02 , //!< readable pattern for 1602 user defined characters
#define ______X_X 0x03 , //!< readable pattern for 1602 user defined characters
#define ____X____ 0x04 , //!< readable pattern for 1602 user defined characters
#define ____X___X 0x05 , //!< readable pattern for 1602 user defined characters
#define ____X_X__ 0x06 , //!< readable pattern for 1602 user defined characters
#define ____X_X_X 0x07 , //!< readable pattern for 1602 user defined characters
#define __X______ 0x08 , //!< readable pattern for 1602 user defined characters
#define __X_____X 0x09 , //!< readable pattern for 1602 user defined characters
#define __X___X__ 0x0a , //!< readable pattern for 1602 user defined characters
#define __X___X_X 0x0b , //!< readable pattern for 1602 user defined characters
#define __X_X____ 0x0c , //!< readable pattern for 1602 user defined characters
#define __X_X___X 0x0d , //!< readable pattern for 1602 user defined characters
#define __X_X_X__ 0x0e , //!< readable pattern for 1602 user defined characters
#define __X_X_X_X 0x0f , //!< readable pattern for 1602 user defined characters
#define X________ 0x10 , //!< readable pattern for 1602 user defined characters
#define X_______X 0x11 , //!< readable pattern for 1602 user defined characters
#define X_____X__ 0x12 , //!< readable pattern for 1602 user defined characters
#define X_____X_X 0x13 , //!< readable pattern for 1602 user defined characters
#define X___X____ 0x14 , //!< readable pattern for 1602 user defined characters
#define X___X___X 0x15 , //!< readable pattern for 1602 user defined characters
#define X___X_X__ 0x16 , //!< readable pattern for 1602 user defined characters
#define X___X_X_X 0x17 , //!< readable pattern for 1602 user defined characters
#define X_X______ 0x18 , //!< readable pattern for 1602 user defined characters
#define X_X_____X 0x19 , //!< readable pattern for 1602 user defined characters
#define X_X___X__ 0x1a , //!< readable pattern for 1602 user defined characters
#define X_X___X_X 0x1b , //!< readable pattern for 1602 user defined characters
#define X_X_X____ 0x1c , //!< readable pattern for 1602 user defined characters
#define X_X_X___X 0x1d , //!< readable pattern for 1602 user defined characters
#define X_X_X_X__ 0x1e , //!< readable pattern for 1602 user defined characters
#define X_X_X_X_X 0x1f , //!< readable pattern for 1602 user defined characters
// above 32 macro definitions generated by Gen1602charPatt.java

/** A DIN resistor symbol (upright) as user defined character pattern. */
extern uint8_t const resistor[];

/** A DIN inductor symbol (upright) as user defined character pattern. */
extern uint8_t const inductor[];

/** A capacitor symbol (upright) as user defined character pattern. */
extern uint8_t const capacitor[];

/** A diode symbol (downward) as user defined character pattern. */
extern uint8_t const diode[];

/** A diode symbol (upward) as user defined character pattern. */
extern uint8_t const diodeUp[];

/** A electrolytic capacitor symbol (+ upward) as character pattern. */
extern uint8_t const elCo[];



//--  helper functions (usually not for application software)   -------------

/** Short (37s) or long (1.52 ms) pause
 *
 *  This is just a helper function not to be used by application software
 *  directly. <br />
 *  Sending command or data to the LCD module TC1602B in succession as in the
 *  initialisation procedere (\ref  initLCD e.g.) requires waits of at
 *  least 37s duration and for some commands even of at least 1520s.
 *  This quite obscure requirement shows the TC1602B
 *  a) being quite slow and
 *  b) having no command buffering at all, regrettably.
 *  The omnipresent requirement for waits is independent of
 *  4 or 8 bit mode. <br />
 *  At least between the first (high) and the second (low) nibble in 4 bit
 *  mode no special wait is needed. Hence, using the 8 bit mode, brings
 *  virtually no advantage. <br />
 *  See the respective manufacturer's data sheet.
 *
 *  So this function is for the >= 37 s and the >= 1520 s recreation. <br />
 *  In simple cases this might be implemented by spin waiting. <br />
 *  In real time automation environments, as \ref intro_sec "weAutSys", this
 *  should be done by the \ref pt "yielding" and \ref timerHdl "timer"
 *  functions featured by the \ref intro_sec "run time". <br />
 *  Extra hardware hint: Considering the TC1602B's timing properties allowing
 *  reading (by putting R/W to a port instead of Gnd) is more helpful than the
 *  8 bit mode (just and only for the purpose of reading the busy/ready bit).
 *
 *  Besides it's bothering operational timing requirements the TC1602B LCD
 *  controller needs 20 respectively 30 ms wait after power up on 5V
 *  respectively 3V supply; this very long delay is not part of this
 *  initialisation helper function. If necessary use \ref preInitLCDpause.
 *
 *  \param loPa true: >= 1520 s  pause; false: >= 37 s pause
 */
void pause4LCDmodule(uint8_t loPa);

/** Set ports used for LCD control as output.
 *
 *  This is just an initialisation helper function. <br />
 *  It sets the operation mode of all digital I/O port bits used as interface
 *  to control the TC1602B LCD controller. <br />
 *  Further it sets the 4 or 8 port bits used as data interface to 0 and
 *  \ref setLCDdataPorts4out() "to output".
 *
 *  Make sure to enable no special functions on any port (bit) used for the
 *  LCD module. <br />
 *  This function would not be called directly by application software.
 *
 *  This is a pure intialisation helper function.
 */
void setLCDcontrolPorts(void);

/** Set ports used for LCD data as output.
 *
 *  This is a pure helper function for initialisation and if feasible
 *  for switching from (data) input back to output.
 *  This function also sets the R/W control line (if given) to 0.
 */
void setLCDdataPorts4out(void);

/** Set ports used for LCD data as input.
 *
 *  This is a pure helper function for switching from (data) output to
 *  input. If the R/W line is \ref RW_NOTFIXED "fixed" to ground this function
 *  does nothing. <br />
 *  Otherwise this function also sets the R/W control line to 1.
 */
void setLCDdataPorts4in(void);


/** Read the address and busy flag
 *
 *  If the R/W on the TC1602B is \ref  RW_NOTFIXED "fixed" to Gnd this
 *  function returns 0xFF. This would signal busy and leed to an endless
 *  loop if (and only if) used very naively in spin waiting for not busy.
 *
 *  This function uses 1.9 .. 2.0s for the 4-bit modes.
 *  \return the busy flag (bit 7) and the current address (cursor position)
 *          (bits 6..0)
 */
uint8_t addressFromLCD(void);
/** Wait for busy flag gone with short (50s) or long (1.6 ms) timeout
 *
 *  Ii the R/W on the TC1602B is \ref  RW_NOTFIXED "fixed" to Gnd this
 *  function will delegate to \ref pause4LCDmodule(loPa) and return the
 *  original parameter value. In those configurations it would be better, i.e.
 *  more self-documenting, to call \ref pause4LCDmodule(loPa) directly.
 *
 *  This function uses spin waiting for reading the address and busy flag
 *  until the busy flag is gone or a timeout. For the background see
 *  \ref pause4LCDmodule.
 *
 *  The return value is the last read. If bit 7, the busy flag, is set, this
 *  function returned because of timeout.
 *
 *  Hint: The timeout values hold for F_CPU 16 MHz. Slower CPUs would
 *  wait longer. This isn't considered a problem as timeout will only
 *  occur on hardware faults or wrong programming. <br />
 *  Usually this function waits as half as long for not busy as would be the
 *  minimal pause according to the data sheet.
 *
 *  If the R/W on the TC1602B is \ref  RW_NOTFIXED "fixed" to Gnd this
 *  function delegates to \ref pause4LCDmodule "pause4LCDmodule(loPa)". The
 *  function \ref pause4LCDmodule will make the specified minimal waits for
 *  any processor clock.
 *
 *  \param loPa true: >= 1.6 ms  timeot; false: >= 50 s timeout
 *  \return last address and busy flag read
 */
uint8_t waitWhileLCDbusy(uint8_t  loPa);


//--  helper functions (not to be called by application software)   ---------

/** Send command (byte) to LCD
 *
 *  This is a helper function, normally not used by application
 *  software. <br />
 *  It sets the command (write) mode and sends the data byte to the LCD
 *  module. It does not include the required command specific pause.
 *
 *  \param command byte (see TC1602B data sheet)
 */
void sendCommand2LCD(uint8_t command);

/** Send 8 bit (full byte) to LCD
 *
 *  Pure helper function, usually not called by application software.
 */
void sendByte2LCD(uint8_t data);

/** Send 4 bit to the LCD module
 *
 *  Pure internal helper function, not to be called by application software.
 *  \param data only the lower nibble is relevant
 */
void send4Bit2LCD(uint8_t data);

/** Send first 4 bit (a mode nibble) to LCD
 *
 *  Pure helper function, not to be called by application software. <br />
 *  And never call it in an 8 bit transfer environment. This may lead to
 *  random behaviour or intentionally raise a compile time error.
 *  \param nib  only the lower nibble is relevant
 */
   void send1st4bitModeNibble(uint8_t nib);

#endif
