/** \file murg.c
 *  An application for easyAVR (7), Arduino and other supported AVR platforms
 *
 *  A little demo program: a clock using the TC1602B LCD display module
 *
 *  This application is to demonstrate the
 -     use of a standard LDC module of the TC1602B family and of
 -     accurate timing by use of an ATmega timer/counter <br />
       as well as hence indirectly the use of interrupts.
 *
 *  Currently we support two hardware target platforms (PLATFORM=..): <br />
 - easyAVR32
 - boAVR2013
 *
 *  These two platforms feature arrays of buttons  &mdash; to be used here to
 *  control clocks and stop-watches &mdash;  and they incorporate the LCD
 *  module in defined neighborhood to the LEDs and buttons.
 *
 *  On two platforms principally supported,
 - arduinoUno and
 - arduMega2560,
 *  these sensors/actuators could be attached also, but not in a pre-defined
 *  arrangement. Hence we recommend just the first two as platform for this
 *  application.
 *
 *  Clean and build by:   <br />
 *    make PLATFORM=easyAVR32    MAIN_F=murg PRG_APP_P=com5 clean all
 *    make PLATFORM=boAVR2013    MAIN_F=murg PRG_APP_P=com14 clean all
 *  depending on platform and flash by
 *    make PLATFORM=easyAVR32    MAIN_F=murg PRG_APP_P=com5  progapp
 *    make PLATFORM=boAVR2013    MAIN_F=murg PRG_APP_P=com14  progapp
 *  setting, of course the right serial port PRG_APP_P for the
 *  serial bootloader.
 *
 *  For the LCD display aspects see all else in the include file
 *  \ref lcd1602b.h. <br />
 *  For pure TC1602B testing and demo use the (sub-) project lahn (see file
 *  \ref lahn.c).
 *
 *  Features:
 -  Wall clock set by buttons: hour +-, minute +-, second +- (start on release)
 -  stop watch controlled by buttons: reset start stop store/recall time
 *
 *  Timebase:
 -    Only for initialisation partly spin waiting for delays / timers,
 -    In the sum / average precise 1ms hardware timer interrupt based
      on the (hopefully) precise Controller clock.
 *
 *
 *  Hint: Up to revision 82 (19.12.2014) this demo application is not yet
 *        ready. I just counts and displays the time since start in the
 *        format bbr />
 *        000 01:23:45.678  (days hours:minutes:seconds.milliseconds). <br />
 *        No button is recognised and ths wall-clock and stoppwatch functions
 *        memtioned above are still to be done.
 *
 *  <br /> &nbsp;
 * \ref copyLicNote " Copyright &copy; " 2014 Albrecht Weinert, Bochum
 * \author  Albrecht Weinert  &nbsp; <<a href="http://a-weinert.de">a-weinert.de</a>>
 * 1
 * 25.01.2017
 */
// Copyright 2014 Albrecht Weinert, Bochum

#include "arch/expConf.h"
// next 3 includes point to elements of weAutSys' serial bootloader project
#include "arch/config.h"
#include "bootloader/boot109.h"
#include "we-aut_sys/ll_common.h"

#include "lcd1602b.h"
#include "formParsUtils.h"
#include "tickTime.h"

#include <avr/pgmspace.h>
#include <stddef.h>

uint8_t resetCauses;

/** murg's welcome greeting and Copyright */
INFLASH(char const mainWlc[]) = "\n\n\n\n\n"
      " * * * *    AVR demo program  murg  V0.3   2014-12-29   * * *  \n"
      "                                                               \n"
      " Demonstration clocks and time on a LCD module (TC1602B).      \n"
      " It links and uses weAutSys' serial bootloader functions.      \n"
      " Copyright (c) 2014  weinert - automation      <weAut.de>      \n"
      " Prof. Dr.-Ing. Albrecht Weinert,  Bochum  <a-weinert.de>      \n"
      "                                                               \n";

ucnt32_t msSinceStart;  // as the name says (implemented)
uint32_t preInitLCDtime = SUPPLY_VOLTAGE == 3 ? 30 : 20;
uint8_t notYetInitLCD = 0xFF;

uint8_t myLast_ms; // helper for consistency / thread safety

// the clock (as yet 0 00:00:00 ist the program start (no controls))
uint8_t d_clock, h_clock, m_clock, s_clock, ts_clock, hs_clock, ms_clock;
#define d_lo line1[2]
#define h_up line1[4]
#define h_lo line1[5]
#define m_up line1[7]
#define m_lo line1[8]
#define s_up line1[10]
#define s_lo line1[11]
#define ts_lo line1[13]
#define hs_lo line1[14]
#define ms_lo line1[15]

//                 0123456789x12345
uint8_t line1[] = "oo0 00:00:00.000@"; // LCD display line 1 (actual content)
uint8_t line2[] = "d+/-h+/-m+/-s+/-@"; // dto. 2

uint8_t beInLine2 = 0;         // display handler / refresher status
uint8_t * inTheLine = line1;  // display handler / refresher status

int main(void) __attribute__((OS_main)); // we are stand alone embedded

/** The application (system) program start
 *
 *  This is the normal system program's program entry.
 *
 *  This application just greets the user on the same serial link (UART) as
 *  used by the serial bootloader.
 *
 *  It then acts (in a loop) as demonstration and testbed for plugable LCD
 *  displays of the TC1602B type or alike.
 */
int main(void){
   // The bootloader leaves reset causes intact if it just returns by
   // timeout. In that case it wasn't really active as bootloader.
   resetCauses = MCUSR; MCUSR = 0; // save & clear reset cause
   ///  basicSystemInit(); // done in bootloader, already

// -- first greeting via serial link (same as  for bootloader)  -----------
   sendSerBytes_P(LOW_ADD(mainWlc));          // a greeting (low flash)
   sendSerBytes_P(FAR_ADD(bootloaderPlatf)); // platform info by bootloader
   sendSerBytes_P(FAR_ADD(bootTextReset));  // bootloader function and text
   sendSerBytes_P(resetCauseText_P(resetCauses));
   sendSerByte('\n'); sendSerByte('\n');

   initSystTiming();
   sei(); // hey, we use an interrupt for 1ms granular timing

   for (;;) { // endless loop
      uint8_t const curr_ms =  tick__1_ms;
      if (curr_ms == myLast_ms) { // inter-tick (do background tasks
         if (notYetInitLCD && msSinceStart.v32 >= preInitLCDtime) { // LCD init
            if (notYetInitLCD & 0x80) { // no basic init
               initLCD(0x08);  // 2line, 5*8, cursor off, no blink
               notYetInitLCD &= ~0x80;
               preInitLCDtime = msSinceStart.v32 + 2; // wait 2ms til next LCD
            } else {
               // last step to LCD ready
               // this is prepared for multistep background LCD init.
               notYetInitLCD = 0; // no further LCD init
            } // further LCD init
            continue;
         }  // LCD init
         // further (non LCD) background tasks
         // (we have none yet)
         continue;
      } // inter-tick (do background tasks)

      // do timing and clock update
      uint8_t msDif = curr_ms - myLast_ms;
      myLast_ms = curr_ms;
      msSinceStart.v32 += msDif;
      ms_clock += msDif;
      if (ms_clock >= 10) {
         ms_clock -= 10;
         if ((++hs_clock) >= 10) {
            hs_clock -= 10;
            if ((++ts_clock) >= 10) {
               ts_clock -= 10;
               if ((++s_clock) >= 60) {
                  s_clock = 0; s_up = s_lo = '0';
                  if ((++m_clock) >= 60) {
                     m_clock = 0; m_up = m_lo = '0';
                     if ((++h_clock) >= 24) {
                        h_clock = 0; h_up = h_lo = '0';
                        ++ d_clock; if (++d_lo > '9') d_lo = '0';
                     } else if (++h_lo > '9') { ++h_up; h_lo = '0'; }
                  } else if (++m_lo > '9') { ++m_up; m_lo = '0'; }
               } else if (++s_lo > '9') { ++s_up; s_lo = '0'; }
            } // ts
            ts_lo = ts_clock + '0';
         } // hs
         hs_lo = hs_clock + '0';
      } // ms
      ms_lo = ms_clock + '0';

      if (notYetInitLCD) continue; // noting else to do yet if no ready LCD

      // refresh one character per millisecond
      switch (beInLine2) {
        case 0: // in line one
           if (inTheLine == &(line1[16])) { // end of line 1
              cursPosLCD(2, 0);
              inTheLine = line2;
              beInLine2 = 1;
              break;
           } // end of line 1
           sendData2LCD(*inTheLine++);
           break;
        case 1: // in line two
           if (inTheLine == &(line2[16])) { // end of line 2
              cursPosLCD(1, 0);
              inTheLine = line1;
              beInLine2 = 0;
              break;
           } // end of line 1
           sendData2LCD(*inTheLine++);
           break;
      } // switch
   } // endless
} // main()  ("murg")
