/** \file neckar.c
 *  An application for easyAVR (7), Arduino and other supported AVR platforms
 *
 *  This simple demo application uses the same USART / serial link as the
 *  serial bootloader to exchange textual messages with a very simple serial
 *  terminal program like vsst.exe or a more comfortable one like HTerm.
 *  Its purpose is to read and set the state of the Controller's ports.
 *
 *  <br /> &nbsp;
 * \ref copyLicNote " Copyright &copy; " 2014 Albrecht Weinert, Bochum
 * \author  Albrecht Weinert  &nbsp; <<a href="http://a-weinert.de">a-weinert.de</a>>
 * $Revision: 1 $
 * $Date: 2017-01-25 11:27:06 +0000 (Wed, 25 Jan 2017) $
 */
// Copyright 2014 Albrecht Weinert, Bochum

#include "arch/expConf.h"
// next 3 includes point to elements of weAutSys' serial bootloader project
#include "arch/config.h"
#include "bootloader/boot109.h"
#include "we-aut_sys/ll_common.h"

#include <avr/pgmspace.h>
#include <avr/interrupt.h>
#include <stddef.h>

uint8_t resetCauses;


/** main's welcome greeting and Copyright */
INFLASH(char const mainWlc[]) = "\n\n\n\n\n\n\n\n\n"
      " * * * *   AVR demo program neckar V0.1               * * * *  \n"
      "                                                               \n"
      " It links and uses weAutSys' serial bootloader functions.      \n"
      " Copyright (c) 2014  weinert - automation      <weAut.de>      \n"
      " Prof. Dr.-Ing. Albrecht Weinert,  Bochum  <a-weinert.de>      \n"
      "                                                               \n";

/** This little demo program's help text. */
INFLASH(char const helpText[]) =
      " Single letter commands to this little demo program:           \n"
      "                                                               \n"
      "  H ? : this help (+ welcome, version and copyright) text      \n"
      "  G   : the greeting (welcome, version and copyright) alone    \n"
      "  B   : enter the bootloader (return after loader's timeout)   \n"
      "  P   : Port, next character chooses a port: A B C D           \n"
      "        third character chooses an operation:                  \n"
      "            O followed by two hex digits = output              \n"
      "            I = input; will reply with x =0xdd<ret>            \n"
      "            D followed by two hex digits = direction           \n"
      "            Have fun playing with ports                        \n";


/** This servers to format hex digits */
INFLASH(char const hexDig[]) = "0123456789ABCDEF ";

/** Parse hex digit to value
 *
 *  @param c the character 0..9, A..F or a..f
 *  @return the hexadecimal value 0..15 resp. 0..0x0F or 0xF0 if illegal c
 */
uint8_t parseHexChar(uint8_t c) {
   if (c >= '0' && c <= '9' ) return c - '0'; // decimal range
   c &= 0xDF; // poor man's uppercase
   if (c >= 'A' && c <= 'F' ) return c - ('A' - 10); // hex digit
   return 0xF0; // c in error
} // parseHexChar(uint8_t)


int main(void) __attribute__((OS_main));

/** The application (system) program start
 *
 *  This is the normal system program's program entry.
 *
 *  At present it is a little bit serial (HMI) communication and a tiny
 *  command line interpreter allowing to control very basic DO/DI port
 *  functions.
 *
 */
int main(void){
   // the bootloader leaves reset causes intact if it just returns by
   // timeout. In that case it wasn't really active as bootloader
   // (communicating with AVRdude e.g.).
   resetCauses = MCUSR; MCUSR = 0; // save & clear reset cause
   ///  basicSystemInit();  // done in boot loader, already
   DDRC  = 0; // set to input (bootloader set to output)
   DDRD  = 0; // set to input (hope does not spoil UART)

   sendSerBytes_P(LOW_ADD(mainWlc));  // a greeting (low flash)
   sendSerBytes_P(FAR_ADD(bootloaderPlatf));  // platform info by bootloader

   sendSerBytes_P(FAR_ADD(bootTextReset));  // bootloader function and text
   sendSerBytes_P(resetCauseText_P(resetCauses));
   sendSerByte('\n'); sendSerByte('\n');

   for(;;) { // endless loop
       if (waitSerByte(0)) { // timeout while waiting 25.6s
           sendSerByte('~'); // yes we can (one AC per 25 s timeout)
           continue; // stay in endless loop
       } // timeout while waiting for UART byte

       uint8_t val = getSerByte(); // Get the byte received (one is ready)
       CLIstartWithGotLetter: // There are bad C habits and elements like
                              // goto. They are outright evil .. and used here.


       // that is the very simple command line interpreter (CLI)
       // one letter = one command (B H G P .. )
       if (val == 'B') {  // B goto bootloader
          bootMain(); // enter the bootloader  (linked bootloader function)
       } // 'B' sends us to the bootloader (until 72s timeout or 'E'  there)

       if (val == 'P' || val == 'p' ) { // P:  port function
          if (waitSerByte(20)) {
             faultyP:
             sendSerBytes("  p {A|B|C..}{i|o|d}[dd] !?! \n\0");
             continue;
          }
          uint8_t port = getSerByte() & 0xDF; // port letter upper case
          sendSerByte(port);  /// TEST
          if (port < 'A' || port > 'G') goto  faultyP;
          if (waitSerByte(20)) goto  faultyP;
          uint8_t poOp = getSerByte() | 0x20; // operation letter lower case
          if (poOp == 'o' || poOp == 'd') { // output or direction register
             uint8_t dig1 = ' ';
             // read value to be output as two hex
             while (dig1 == ' ') dig1 = recvSerByte(); // ignore leading space
             uint8_t dig2 = recvSerByte();
             uint8_t outVal = parseHexChar(dig2);
             if (outVal == 0xF0)  goto  faultyP;
             uint8_t dig1val = parseHexChar(dig1);
             if (dig1val == 0xF0)  goto  faultyP;
             outVal |= (dig1val << 4);

             if (poOp == 'o') {
                sendSerBytes("\nwriting to port"); sendSerByte(port);
                sendSerBytes(" 0x"); sendSerByte(dig1); sendSerByte(dig2);
                sendSerBytes(" \n\0");
                switch(port) {
                   #ifdef PORTA
                   case 'A' : PORTA = outVal; break;
                   #endif
                   case 'B' : PORTB = outVal; break;
                   case 'C' : PORTC = outVal; break;
#ifdef PORTD
                   case 'D' : PORTD = outVal; break;
#endif
                #ifdef PORTE
                   case 'E' : PORTE = outVal; break;
                #endif
                #ifdef PORTF
                   case 'F' : PORTF = outVal; break;
                #endif
                #ifdef PORTG
                   case 'G' : PORTG = outVal; break;
                #endif
                } // switch
                continue;
             }  // poOp == 'o'

             // to here with poOp == 'd'
             sendSerBytes("\nwriting to port"); sendSerByte(port);
             sendSerBytes("'s direction register 0x");
             sendSerByte(dig1); sendSerByte(dig2); sendSerBytes(" \n\0");
             switch(port) {
                #ifdef PORTA
                case 'A' : DDRA = outVal; break;
                #endif
                case 'B' : DDRB = outVal; break;
                case 'C' : DDRC = outVal; break;
#ifdef PORTD
                case 'D' : DDRD = outVal; break;
#endif
             #ifdef DDRE
                case 'E' : DDRE = outVal; break;
             #endif
             #ifdef DDRF
                case 'F' : DDRF = outVal; break;
             #endif
             #ifdef DDRG
                case 'G' : DDRG = outVal; break;
             #endif
             } // switch
             continue; // ready direction register write

          }
          if (poOp != 'i') goto  faultyP; // only d i o are legal
              //          123456789.12                 3     456
          sendSerBytes("\nreading port"); sendSerByte(port);
              //        45
          sendSerBytes("= ");
          uint8_t inVal = 0;
          for(;;++inVal) {
                     //   1234567.9
            sendSerBytes(" 0x");
            switch(port) {
#ifdef PINA
             case 'A' : inVal = PINA; break;
#endif
             case 'B' : inVal = PINB; break;
             case 'C' : inVal = PINC; break;
#ifdef PIND
             case 'D' : inVal = PIND; break;
#endif
#ifdef PINE
             case 'E' : inVal = PINE; break;
#endif
#ifdef PINF
             case 'F' : inVal = PINF; break;
#endif
#ifdef PING
             case 'G' : inVal = PING; break;
#endif
             default  : --inVal; break; // decrementing shows error
            }
            sendSerByte(  pgm_read_byte(& hexDig[inVal >> 4]  ) );
            sendSerByte(  pgm_read_byte(& hexDig[inVal & 0x0F]) );
            if (! waitSerByte(10)) { // wait 1 s or for RdX
               val = getSerByte();
               if (val <  ' ') continue; //swallow own LF or empty lines
               if (val <  'A') break;  // swallow non letter but break input
               goto CLIstartWithGotLetter; // allow letter read as next command
            }
          } // for

          sendSerBytes(" \n\0");
          continue;
       } // 'P' port (not ready)


       if (val == 'F') { // send FF serially (until key received)
          for (;;) { // for endless 0xFF
             sendSerByte(0xFF);
             if (isSerByteRec()) { // ser in
                val = getSerByte();
                if (val == CR || val == LF) continue; // for sake of HTerm
                break; // for endless 0xFF
             } // // ser in
          } // endless 0xFF
          sendSerBytes(" \n * * end sending 0xFF... \n\n \0");
          continue;
       } // 'F' is serial link test

       val &= 0xDF; // command letter upper case
       if (val == 'H' || val == 'G' || val == '?') { // ? H Help (Hilfe)
          sendSerBytes_P(LOW_ADD(mainWlc));  // a greeting (low flash)
          sendSerBytes_P(FAR_ADD(bootloaderPlatf));  // platform info
          if ( val == 'G' ) continue;
          sendSerByte('\n');
          sendSerBytes_P(LOW_ADD(helpText));
          sendSerByte('\n'); sendSerByte('\n');
          continue;
       } // end CLI help

       // end of CLI interpreted commands all else is just echoed
       sendSerByte(val); // All else character are echoed by this "CLI"
   } // endless for

} // main()


